@@ -5,13 +5,16 @@
 
 from __future__ import annotations
 
+import ast
+import re
 import textwrap
+from unittest import mock
 
 import pytest
 
 from coverage import env
-from coverage.exceptions import NotPython
-from coverage.parser import PythonParser
+from coverage.exceptions import NoSource, NotPython
+from coverage.parser import PythonParser, is_constant_test_expr
 
 from tests.coveragetest import CoverageTest
 from tests.helpers import arcz_to_arcs
@@ -120,7 +123,7 @@ class PythonParserTest(PythonParserTestBase):
         )
     ])
     def test_not_python(self, text: str) -> None:
-        msg = r"Couldn't parse '<code>' as Python source: '.*' at line \d+"
+        msg = r"Couldn't parse '<code>' as Python source: ['\"].*['\"] at line \d+"
         with pytest.raises(NotPython, match=msg):
             _ = self.parse_text(text)
 
@@ -195,6 +198,24 @@ class PythonParserTest(PythonParserTestBase):
         with pytest.raises(NotPython, match=msg):
             self.parse_text("]")
 
+    def test_bug_1891(self) -> None:
+        # These examples exercise code paths I thought were impossible.
+        parser = self.parse_text("""\
+            res = siblings(
+                'configure',
+                **ca,
+            )
+            """)
+        assert parser.exit_counts() == {1: 1}
+        parser = self.parse_text("""\
+            def g2():
+                try:
+                    return 2
+                finally:
+                    return 3
+            """)
+        assert parser.exit_counts() == {1: 1, 2: 1, 3: 1, 5: 1}
+
 
 class ExclusionParserTest(PythonParserTestBase):
     """Tests for the exclusion code in PythonParser."""
@@ -673,6 +694,19 @@ class ExclusionParserTest(PythonParserTestBase):
         )
         assert parser.statements == {2, 4}
 
+    def test_no_exclude_at_all(self) -> None:
+        parser = self.parse_text("""\
+            def foo():
+                if fooey:
+                    a = 3
+                else:
+                    a = 5
+            b = 6
+            """,
+            exclude="",
+        )
+        assert parser.exit_counts() == { 1:1, 2:2, 3:1, 5:1, 6:1 }
+
     def test_formfeed(self) -> None:
         # https://github.com/nedbat/coveragepy/issues/461
         parser = self.parse_text("""\
@@ -1068,6 +1102,24 @@ class MatchCaseMissingArcDescriptionTest(PythonParserTestBase):
         # 4-6 isn't a possible arc, so the description is generic.
         assert parser.missing_arc_description(4, 6) == "line 4 didn't jump to line 6"
 
+    def test_missing_arc_descriptions_bug1775(self) -> None:
+        # Bug: the `if x == 2:` line was marked partial with a message of:
+        # line 6 didn't return from function 'func', because the return on line 4 wasn't executed
+        # At some point it changed to "didn't jump to the function exit," which
+        # is close enough. These situations are hard to describe precisely.
+        parser = self.parse_text("""\
+            def func():
+                x = 2
+                try:
+                    return 4
+                finally:
+                    if x == 2:  # line 6
+                        print("x is 2")
+
+            func()
+            """)
+        assert parser.missing_arc_description(6, -1) == "line 6 didn't jump to the function exit"
+
 
 class ParserFileTest(CoverageTest):
     """Tests for coverage.py's code parsing from files."""
@@ -1134,3 +1186,33 @@ class ParserFileTest(CoverageTest):
 
         parser = self.parse_file("abrupt.py")
         assert parser.statements == {1}
+
+    def test_os_error(self) -> None:
+        self.make_file("cant-read.py", "BOOM!")
+        msg = "No source for code: 'cant-read.py': Fake!"
+        with pytest.raises(NoSource, match=re.escape(msg)):
+            with mock.patch("coverage.python.read_python_source", side_effect=OSError("Fake!")):
+                PythonParser(filename="cant-read.py")
+
+
+@pytest.mark.parametrize(
+    ["expr", "ret"],
+    [
+        ("True", (True, True)),
+        ("False", (True, False)),
+        ("1", (True, True)),
+        ("0", (True, False)),
+        ("__debug__", (True, True)),
+        ("not __debug__", (True, False)),
+        ("not(__debug__)", (True, False)),
+        ("-__debug__", (False, False)),
+        ("__debug__ or True", (True, True)),
+        ("__debug__ + True", (False, False)),
+        ("x", (False, False)),
+        ("__debug__ or debug", (False, False)),
+    ]
+)
+def test_is_constant_test_expr(expr: str, ret: tuple[bool, bool]) -> None:
+    node = ast.parse(expr, mode="eval").body
+    print(ast.dump(node, indent=4))
+    assert is_constant_test_expr(node) == ret
